#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#ifdef __BORLANDC__
#include <io.h>
#else
#include <unistd.h>
#endif

#define EKALABEL "S"
#define ENDLABEL "Q"

static int flagcpp = 0;

static void Usage(void)
{
    printf(
        "Usage: TTTC [-nologo] [-cpp] infile[.ttt] [outfile[.c]]\n\n"
        "Compiles ttt conversations to c files.\n"
		"See Project Notes in tttc.prj for more details.\n");
    exit(0);
}

static FILE *f, *t;

static char ModName[16];

static const char *CStr(const char *s)
{
	static char LineBuf[2048];
	char *q;
	for(q=LineBuf; *s; s++)
    {
		if(*s=='"')*q++='\\';
		if(*s!='\n' && *s!='\r')*q++ = *s;
    }
    *q=0;
    return LineBuf;
}

static void Compile(void)
{
	char LineBuf[2048];
	int Line, ekac;
	char Label[64];
	long l;
	
	fprintf(t,
        "#ifdef __DJ_size_t\n"
        "#define _SIZE_T\n"
        "#endif\n"
		"#ifndef _SIZE_T\n"
		"#define tttcfull\t\t/* Kind of bad test... */\n");
		
	if(flagcpp)
		fprintf(t,
			"#endif\n\n"
			"#ifndef tttobj_def\n"
			"#define tttobj_def\n"
			"#include <vector>\n"
			"#include <string>\n\n"
			"using namespace std;\n"
			"class tttobj\n"
			"{\n"
			"public:\n"
			"\tint id;\n"
			"\tstring to, text;\n"
			"\ttttobj(const int ID,const char *TO,const char *TEXT)\n"
			"\t:\tid(ID), to(TO), text(TEXT)\n"
			"\t{\n"
			"\t}\n"
			"};\n"
			"typedef vector <tttobj> tttlist;\n"
			"\n"
			"extern const tttobj &DoQuery(const tttlist &s);\n"
			"extern bool TTTDead(void);\n"
			"\n"
			"#endif\n"
			"\n"
			"#ifdef tttcfull\n");
	else
		fprintf(t,
			"\n"
			"#include <stdio.h>\n"
			"#include <stdlib.h>\n"
			"#include <string.h>\n"
			"\n"
			"#include \"ttt.h\"\n"
			"\n"
			"static const char *IniPos=\"" EKALABEL "\";\n"
			"static const char *ThisPos;\n"
			"static int SelectionCount;\n"
			"static const char *Selections[32];\n"
			"static const char *SelectDest[32];\n");

	puts("Pass 1: Compiling the '#' -lines");

	clearerr(f);
	for(rewind(f), ekac=Line=1; !ferror(t); Line++)
	{
		char *q;
		if(!fgets(LineBuf, sizeof(LineBuf)-1, f))
		{
			printf("At line %d: Unexpected end of file\n", Line);
			goto P5;
		}
		if(LineBuf[0]=='>' || LineBuf[0]==':')goto P5;

		for(q=LineBuf; isspace(*q); q++);
		if(*q=='#')
		{
			if(ekac)
			{
				ekac=0;
				fprintf(t, "\n");
			}
			fprintf(t, "%s", q+1);
		}
	}
	goto OutFErr;

P5: if(flagcpp)
		fprintf(t,
			"class ttthelper\n"
			"{\n"
			"public:\n"
			"\tstring key;\n"
			"\ttttobj obj;\n"
			"\tttthelper(const string &k, int id, const char *dest,\n"
			"\t          const char *text):key(k),obj(id,dest,text)\n"
			"\t{\n"
			"\t}\n"
			"};\n"
			"typedef vector<ttthelper> tttentry;\n"
			"static void add(tttentry &e, const char *key, int id,\n"
			"                const char *dest, const char *text)\n"
			"{\n"
			"\te.insert(e.end(), ttthelper(key, id, dest, text));\n"
			"}\n");
	fprintf(t,
        "#else\n"
        "extern\n"
		"#endif\n\n"
		"void %s_ttt(void)\n"
		"#ifdef tttcfull\n"
		"{\n", ModName);
		
	if(flagcpp)
		fprintf(t,
			"\ttttentry selections;\n"
			"\tstring ThisPos=\"" EKALABEL "\";\n\n");
	else
		fprintf(t,
			"\tint ttt_Eka;\n"
			"\tThisPos=IniPos;\n"
			"\tLastSaid=\"\";\n"
			"\tif(WasSaid)*WasSaid=0;\n"
			"\tfor(ttt_Eka=1; strcmp(ThisPos, \"" ENDLABEL "\"); ttt_Eka=0)"
			"\n\t{\n");
			
	if(flagcpp)
	{
		puts("Pass 2: Compiling ':' and '>' lines");
		clearerr(f);
		for(rewind(f), ekac=1, Line=0; !ferror(t); )
		{
			int id, c;
			char ideet[] = ":>";
			char *s, *q;
			
			if(!fgets(LineBuf, sizeof(LineBuf)-1, f))goto P4;
		P8:	Line++;
			id=-1;	/* keep gcc happy. */
			if(LineBuf[0]==ideet[0])id=0;
			else if(LineBuf[0]==ideet[1])id=1;
			else
			{
				if(LineBuf[0]=='#')fprintf(t, "\t%s", LineBuf+1);
				continue;
			}
			
			for(s=LineBuf; *++s!=ideet[id]; )
			{
				if(*s==',')
					printf("At line %d: Warning: ',' is not yet supported in C++ mode.\n", Line);
				if(!*s||*s=='\n')
				{
					printf("At line %d: Expected '%c'\n", Line, ideet[id]);
					goto Vihre;
				}
			}
			*s=0;
			for(q=++s; *s!=ideet[id]; s++)
				if(!*s||*s=='\n')
				{
					printf("At line %d: Expected '%c'\n", Line, ideet[id]);
					goto Vihre;
				}
			fprintf(t, "\tadd(selections, \"%s\"", CStr(LineBuf+1));
			
			*s = 0;
			fprintf(t, ", %d, \"%s\", \"", id, CStr(q));
			
            for(q=++s, c=-1; ; Line++)
            {
				while(isspace(*q))q++;
				if(*q!='/' || q[1]!='/')
				{
					if(*q=='#')
					{
						/* Warning: '#' works a bit differently in C++ mode. */
						break;
					}					
					if(!c)fprintf(t, "\\n");
					fprintf(t, "%s", CStr(q));
					c=0;
				}
RepFGets3:		if(!fgets(LineBuf, sizeof(LineBuf)-1, f)){c=1;break;}
				q=LineBuf;while(isspace(*q))q++;if(!*q)goto RepFGets3;
				if(strchr(ideet, LineBuf[0]))break;
				q=LineBuf;
			}
			fprintf(t, "\");\n");
			if(c==1)goto P4;
			goto P8;
		}
		goto OutFErr;
	}
	
	puts("Pass 2: Compiling ':' -lines");

	clearerr(f);
	for(rewind(f), ekac=1, Line=0; !ferror(t); )
	{
		if(!fgets(LineBuf, sizeof(LineBuf)-1, f))goto P1;
P2:     Line++;

		if(LineBuf[0]==':')
		{
			char *s, *q;
			int c;
			
			if(!flagcpp)
				fprintf(t, "\t\t%sif(", ekac?ekac=0,"":"else ");
			
			for(q=LineBuf; ; )
			{
				for(s=++q; *s!=':' && *s!=','; s++)
					if(!*s)
					{
						printf("At line %d: Expected ':' or ','\n", Line);
						goto Vihre;
					}
				c=*s;
				*s=0;
				
				if(flagcpp)
					fprintf(t, "ThisPos==\"%s\"", CStr(q));
				else
					fprintf(t, "!strcmp(ThisPos, \"%s\")", CStr(q));				
				
                if(c==':')break;
				fprintf(t, "||");
				printf("At line %d: Warning: It is ineffecient to use ',' with ':'\n", Line);
                q=s;
			}

			fprintf(t, ")\n\t\t{\n");
			for(q=++s; *s!=':'; s++)
				if(!*s)
				{
					printf("At line %d: Expected ':'\n", Line);
					goto Vihre;
				}
            *s=0;
			fprintf(t, "\t\t\tThisPos=\"%s\";\n", CStr(q));

            for(q=++s, c=0; ; Line++)
            {
				while(isspace(*q))q++;
				if(*q!='/' || q[1]!='/')
				{
					if(*q=='#')
                        fprintf(t, "\t\t\t%s", q+1);
					else
						fprintf(t, "\t\t\tSay(\"%s\");\n", CStr(q));
				}

RepFGets1:		if(!fgets(LineBuf, sizeof(LineBuf)-1, f)){c=1;break;}
				q=LineBuf;while(isspace(*q))q++;if(!*q)goto RepFGets1;
				if(LineBuf[0]==':' || LineBuf[0]=='>')break;
				q=LineBuf;
			}
			fprintf(t, "\t\t}\n");
			if(c==1)goto P1;
			goto P2;
		}
	}
OutFErr:
	perror("Outfile error");
Vihre:
	fclose(f);
	fclose(t);
	exit(errno);

P1:	if(!ekac)
	{
		fprintf(t,
			"\t\telse if(!ttt_Eka) /* We don't need to warn for first text (IniPos) */\n");
		
		if(flagcpp)
			fprintf(t,
				"\t\t\tSay(\"Internal error - missing TTT reply :\" + ThisPos + \":.\");\n");
		else
			fprintf(t,
				"\t\t{\n"
				"\t\t\tchar Buff[256];\n"
				"\t\t\tsprintf(Buff, \"Internal error - missing TTT reply :%%s:.\", ThisPos);\n"
				"\t\t\tSay(Buff);\n"
				"\t\t}\n");
	}

	puts("Pass 3: Compiling '>' -lines");

	fprintf(t,
		"\n\t\tif(");
		
	if(flagcpp)
		fprintf(t, "ThisPos == \"" ENDLABEL "\"");
	else
		fprintf(t, "!strcmp(ThisPos, \"" ENDLABEL "\")");
		
	fprintf(t,
		")break;\n"
		"\n\t\tSelectionCount=0;\n\n");

	clearerr(f);
	for(rewind(f), Line=0; !ferror(t); )
	{
		if(!fgets(LineBuf, sizeof(LineBuf)-1, f))goto P4;
P3:     Line++;

		if(LineBuf[0]=='>')
		{
			char *s, *q;
			char Ehto[32];
			int c;
			for(fprintf(t, "\t\tif("), q=LineBuf; ; )
			{
				for(s=++q; *s!='>' && *s!=','; s++)
					if(!*s)
					{
						printf("At line %d: Expected '>' or ','\n", Line);
						goto Vihre;
					}
				c=*s;
				*s=0;
				fprintf(t, "!strcmp(ThisPos, \"%s\")", CStr(q));
				if(c=='>')break;
				*s=c;
				fprintf(t, "||");
				q=s;
			}

			strcpy(Ehto, LineBuf);

			fprintf(t, ")\n\t\t{\n");
Plus:       for(q=++s; *s!='>'; s++)
				if(!*s)
				{
					printf("At line %d: Expected '>'\n", Line);
					goto Vihre;
				}
			*s=0;
			fprintf(t, "\t\t\tSelectDest[SelectionCount]=\"%s\";\n", CStr(q));

			for(q=++s, c=0; ; Line++)
			{
				while(isspace(*q))q++;
				if(*q!='/' || q[1]!='/')
				{
					if(*q=='#')
						fprintf(t, "\t\t\t%s", q+1);
					else
					{
						if(c)
							printf("At line %d: Error: '>' lines should have ONLY one text line.\n"
								   "            More text lines cause malfunction.\n", Line);

						fprintf(t, "\t\t\tSelections[SelectionCount]=\"%s\";\n", CStr(q));
						c=2;
					}
				}

RepFGets2:		if(!fgets(LineBuf, sizeof(LineBuf)-1, f)){c=1;break;}
				q=LineBuf;while(isspace(*q))q++;if(!*q)goto RepFGets2;
				if(LineBuf[0]==':')break;
				if(LineBuf[0]=='>')
				{
					if(strncmp(LineBuf, Ehto, strlen(Ehto)))break;
					q = LineBuf+strlen(Ehto);
					if(*q != '>')break;
					fprintf(t, "\t\t\tSelectionCount++;\n");
					s=q;
					goto Plus;
				}
				q=LineBuf;
			}
			fprintf(t, "\t\t\tSelectionCount++;\n\t\t}\n");
			if(c==1)goto P4;
			goto P3;
		}
	}
	goto OutFErr;

P4: if(flagcpp)
		fprintf(t,
			"\n\tint id=0;\n"
			"\n\twhile(ThisPos != \"" ENDLABEL "\" && !TTTDead())\n"
			"\t{\n"
			"\t\tbool idset=false; tttlist l; tttentry::const_iterator i;\n\n"
			"\t\tfor(i=selections.begin(); i!=selections.end(); i++)\n"
			"\t\t\tif(i->key==ThisPos)\n"
			"\t\t\t{\n"
			"\t\t\t\tif(idset && id != i->obj.id)\n"
			"\t\t\t\t{\n"
			"\t\t\t\t\tl.clear();\n"
			"\t\t\t\t\tl.insert(l.end(), tttobj(id, \""ENDLABEL"\",\n"
			"\t\t\t\t\t\tstring(\"ttt error: id mismatch at \"+ThisPos).c_str()));\n"
			"\t\t\t\t\tbreak;\n"
			"\t\t\t\t}\n"
			"\t\t\t\tidset=true, id=i->obj.id;\n"
			"\t\t\t\tl.insert(l.end(), i->obj);\n"
			"\t\t\t}\n"
			"\t\tif(l.empty())\n"
			"\t\t\tl.insert(l.end(), tttobj(id, \""ENDLABEL"\",\n"
			"\t\t\t\tstring(\"ttt error: no reply for \"+ThisPos).c_str()));\n\n"
			"\t\tThisPos=DoQuery(l).to;");
	else
		fprintf(t,
			"\t\tif(SelectionCount)\n"
			"\t\t{\n"
			"\t\t\tSelections[SelectionCount]=NULL;\n"
			"\t\t\tThisPos=SelectDest[DoQuery(SelectionCount, Selections)];\n"
			"\t\t\tif(WasSaid)*WasSaid=0;\n"
			"\t\t}");
			
	fprintf(t,
		"\n\t}\n"
		"}\n"
		"#undef tttcfull\n"
		"#else\n"
		";\n"
		"#endif\n");

	if(flagcpp)
		puts("Pass 3: Checking for errors.");
	else
		puts("Pass 4: Checking for errors.");

	/* Thn pisteeseen psty tiedoston pitisi */
	/* olla puhdas ':' ja '>' -rivien osalta.     */

	clearerr(f);
	strcpy(Label, EKALABEL);
	rewind(f);

	for(l=Line=0, ekac=1; !ferror(t); )
	{
		char *q, *s;
		int c=0;

		for(rewind(f);;)
		{
			if(!fgets(LineBuf, sizeof(LineBuf)-1, f))break;
			if(LineBuf[0]!='>' && LineBuf[0]!=':')continue;

			for(q=LineBuf; ; )
			{
				for(s=++q; *s!=':' && *s!='>' && *s!=','; s++)
					if(!*s)
					{
						printf("At line %d: Expected ':' or ','\n", Line);
						goto Vihre;
					}
				c=*s;
				*s=0;
				if(!strcmp(q, Label))goto P7;
				if(c!=',')break;
				q=s;
			}
		}

		if(Line)printf("At line %d", Line);else printf("Global");
		printf(": Warning! There exists NO entry for label \"%s\"!\n", Label);

P7:		for(fseek(f,l,SEEK_SET); !ferror(t); )
		{
			if(!fgets(LineBuf, sizeof(LineBuf)-1, f))goto P6;
			Line++;
			if(LineBuf[0]=='>' || LineBuf[0]==':')break;
		}

		ekac=0;

		c = LineBuf[0];
		for(s=LineBuf+1; *s!=c; s++);
		for(q=++s; *s!=c; s++);
		*s = 0;
		strcpy(Label, q);
		l=ftell(f);

		if(!strcmp(Label, ENDLABEL))goto P7;
	}
	goto OutFErr;

P6:;
}

int main(int argc, const char *const *argv)
{
	char in[256];
	char out[256];
	char *s, *q;
	int a;

	for(a=1; !strcmp(argv[1]?argv[1]:"", "-nologo"); ++argv, --argc)a=0;
	for(; !strcmp(argv[1]?argv[1]:"", "-cpp"); ++argv, --argc)flagcpp=1;

	if(a)
		puts(
            "TTTC compiler Copyright (C) 1992,1999 Bisqwit (bisqwit@iki.fi)\n"
			"  29.7.1998: Basic version release\n"
			"  31.7.1998: Added optimizations\n"
			"             Added support for empty lines\n"
			"             Added support for '#'-lines in header\n"
			"   1.8.1998: Added pass 4: warnings.\n"
			"  3.10.1998: Separate project.\n"
			"  28.9.1999: Posix, Unix-support.\n"
			"  8.10.1999: Added -cpp switch.\n"
			" 12.10.1999: Rewrote -cpp code.\n");
			
	if(flagcpp)
		puts("C++ mode selected\n");

	if(!--argc)Usage();
	//fprintf(stderr, "argv[1]='%s' (%d)\n", argv[1], strlen(argv[1]));
	strcpy(in, argv[1]);
	//fprintf(stderr, "strcpy oli ok\n");
	out[0]=0;
	if(argc>0)
	{
		if(--argc)strcpy(out, argv[2]);
		if(argc>1)Usage();
	}

	s = strrchr(in, '.');
	q = strrchr(in, '\\');
	if(!s || s < q)strcat(in, ".ttt");
	s = strrchr(in, '.');
	q = strrchr(in, '\\');
	if(!q)q=in;
	if(s-q >= (int)sizeof(ModName))
	{
		printf("TTTC: Internal error - Overflow, %d bytes\n", s-q);
        return (int)(s-q);
    }
	strncpy(ModName, q, (int)(s-q));

    if(out[0])
    {
        s = strrchr(out, '.');
        q = strrchr(out, '\\');
        if(!s || s < q)
        	strcat(out, flagcpp ? ".cc" : ".c");
    }
    else
    {
        dup2(2, 1);
        strcpy(out, in);
        s = strrchr(out, '.');
        strcpy(s, flagcpp ? ".cc" : ".c");
    }

    f = fopen(in, "rt");
    if(!f)
    {
        puts("Can't open infile");
        perror(in);
        exit(errno);
    }
    t = fopen(out, "w");
    if(!t)
    {
        puts("Can't open outfile");
        perror(out);
        exit(errno);
    }

    Compile();

    printf("Done - written %s\n", out);

    fclose(f);
    fclose(t);

    return 0;
}
